from __future__ import annotations

from dataclasses import dataclass, field
from itertools import chain

from pdm.exceptions import CandidateNotFound
from resolvelib.resolvers import (
    RequirementInformation,
    Resolution,
    ResolutionImpossible, Resolver,
    _build_result,
)

from pdm_conda import logger
from pdm_conda.conda import CondaResolutionError
from pdm_conda.environments import CondaEnvironment
from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import CondaRequirement, as_conda_requirement

CONSTRAINS_KEY = "___constrains___"
CONDA_RESOLUTION_KEY = "___conda_resolution___"


@dataclass
class State:
    mapping: dict
    criteria: dict
    backtrack_causes: list
    constrains: dict = field(default_factory=dict)
    conda_resolution: dict = field(default_factory=dict)


class CondaResolution(Resolution):
    def __init__(
        self,
        *args,
        base_constrains: dict | None = None,
        conda_resolution: dict | None = None,
        is_conda_environment: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._is_conda_environment = is_conda_environment
        self._base_constrains = base_constrains or dict()
        self._conda_resolution = conda_resolution or dict()

    @property
    def state(self):
        if self._states and not isinstance(state := self._states[-1], State):
            self._states[-1] = State(
                state.mapping,
                state.criteria,
                state.backtrack_causes,
                self._base_constrains,
                self._conda_resolution,
            )
        return super().state

    def _push_new_state(self):
        base = self.state
        base.constrains = base.criteria.pop(CONSTRAINS_KEY, base.constrains)
        base.conda_resolution = base.criteria.pop(CONDA_RESOLUTION_KEY, base.conda_resolution)
        state = State(
            mapping=base.mapping.copy(),
            criteria=base.criteria.copy(),
            backtrack_causes=base.backtrack_causes[:],
            constrains=base.constrains.copy(),
            conda_resolution=base.conda_resolution.copy(),
        )
        # update repository conda resolution to latest
        if self._is_conda_environment:
            self._p.repository.update_conda_resolution(resolution=state.conda_resolution)
        self._states.append(state)

    def _remove_information_from_criteria(self, criteria, parents):
        state = self.state
        state.constrains = state.criteria.pop(CONSTRAINS_KEY, state.constrains)
        state.conda_resolution = state.criteria.pop(CONDA_RESOLUTION_KEY, state.conda_resolution)
        super()._remove_information_from_criteria(criteria, parents)

    def _ensure_criteria(self, criteria):
        # save temporarily constrains and conda_resolution in criteria
        for k, v in [(CONDA_RESOLUTION_KEY, self.state.conda_resolution), (CONSTRAINS_KEY, self.state.constrains)]:
            if k not in criteria:
                criteria[k] = v.copy()

    def update_constrains(self, candidate: CondaCandidate, criteria=None, merge_old: bool | set[str] = True):
        if isinstance(merge_old, bool):
            if merge_old:
                merge_old = set(candidate.constrains.keys())
            else:
                merge_old = set()
        if criteria is not None:
            self._ensure_criteria(criteria)
            constrains = criteria[CONSTRAINS_KEY]
        else:
            constrains = self._base_constrains

        for identifier, constrain in candidate.constrains.items():
            if identifier != "python":
                # keep most restrictive constrain
                if identifier in merge_old and (existing_constrain := constrains.get(identifier, None)) is not None:
                    constrain = existing_constrain.merge(constrain)
                constrains[identifier] = constrain
                if criteria is not None and identifier in criteria:
                    # todo msg
                    self._add_to_criteria(criteria, constrain, parent=candidate)

    def _update_conda_resolution(self, criteria, parent):
        requirements = list(
            chain.from_iterable(
                (
                    [information.requirement for information in criterion.information]
                    for i, criterion in criteria.items()
                    if i not in (CONDA_RESOLUTION_KEY, CONSTRAINS_KEY) and criterion.information
                )
            ),
        )
        if parent is not None:
            requirements.extend(self._p.get_dependencies(candidate=parent))
        self._ensure_criteria(criteria)
        self._p.repository.update_conda_resolution(
            [self._p.get_requirement_from_overrides(req) for req in requirements],
            criteria[CONDA_RESOLUTION_KEY],
        )

    def _add_to_criteria(self, criteria, requirement, parent):
        _req = requirement
        if self._is_conda_environment:
            # merge with constrain if exists
            self._ensure_criteria(criteria)
            constrains = criteria[CONSTRAINS_KEY]
            if (constrain := constrains.get(requirement.conda_name, None)) is not None:
                _req = constrain.merge(requirement)

            identifier = self._p.identify(_req)
            if criterion := criteria.get(identifier):
                excluded = self._p.repository.environment.project.conda_config.excluded_identifiers
                if isinstance(_req, CondaRequirement):
                    # if conda requirement but other not conda requirement and excluded
                    # then transform to named requirement
                    if any(
                            not isinstance(i.requirement, CondaRequirement) and i.requirement.identify() in excluded
                            for i in criterion.information
                    ):
                        _req = _req.as_named_requirement()
                    # else any other to conda
                    else:
                        criterion.information = [
                            RequirementInformation(as_conda_requirement(i.requirement), i.parent)
                            for i in criterion.information
                        ]

                # if excluded then delete conda related information else if other conda requirement transform to conda
                else:
                    if requirement.identify() in excluded:
                        criterion.information = [
                            (
                                RequirementInformation(i.requirement.as_named_requirement(), i.parent)
                                if isinstance(
                                    i.requirement,
                                    CondaRequirement,
                                )
                                else i
                            )
                            for i in criterion.information
                        ]
                    elif any(isinstance(i.requirement, CondaRequirement) for i in criterion.information):
                        _req = as_conda_requirement(requirement)
            try:
                self._update_conda_resolution(criteria, parent)
            except CondaResolutionError as err:
                logger.info(err)
                raise CandidateNotFound(f"Unable to find candidates for "
                                        f"[success]{_req.name}[/success] with Conda.") from err
        super()._add_to_criteria(criteria, _req, parent)

    def _get_updated_criteria(self, candidate):
        criteria = super()._get_updated_criteria(candidate)
        # merge with previous constrain if exists
        if isinstance(candidate, CondaCandidate) and self._is_conda_environment:
            self.update_constrains(candidate, criteria)
        return criteria

    def initialize_conda_resolution(self, requirements):
        # update conda resolution
        self._p.repository.update_conda_resolution(
            [self._p.get_requirement_from_overrides(req) for req in requirements],
            self._conda_resolution,
        )
        # update constrains
        for candidates in self._conda_resolution.values():
            for can in candidates:
                self.update_constrains(candidate=can)


class CondaResolver(Resolver):
    def resolve(self, requirements, max_rounds=100):
        is_conda_environment = isinstance(self.provider.repository.environment, CondaEnvironment)
        resolution = CondaResolution(self.provider, self.reporter, is_conda_environment=is_conda_environment)
        if is_conda_environment:
            project = self.provider.repository.environment.project
            conda_config = project.conda_config
            resolution.initialize_conda_resolution(requirements)
            if conda_config.is_initialized and conda_config.custom_behavior:
                project.is_distribution = True
        else:
            assert not any(isinstance(r, CondaRequirement) for r in requirements)

        try:
            state = resolution.resolve(requirements, max_rounds=max_rounds)
            result = _build_result(state)
        finally:
            if is_conda_environment:
                project.is_distribution = None

        # here we remove a self dependency we added because of the custom behavior
        if is_conda_environment:
            if conda_config.is_initialized and conda_config.custom_behavior and not project.is_distribution:
                for key, candidate in list(result.mapping.items()):
                    if candidate.name == conda_config.project_name:
                        del result.mapping[key]
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.update import Command as BaseCommand
from pdm.models.specifiers import get_specifier

from pdm_conda.models.requirements import CondaRequirement
from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "update"

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        super().handle(project=project, options=options)
        project = cast(CondaProject, project)
        if project.conda_config.is_initialized and project.conda_config.custom_behavior:
            if not options.dry_run and options.save_strategy:
                groups = set(project.iter_groups(dev=False))
                dev_groups = set(project.iter_groups(dev=True)) - groups
                candidates = project.locked_repository.all_candidates
                requirements = dict()
                for i, group in enumerate(groups | dev_groups):
                    group_requirements = dict()
                    for identifier, req in project.get_dependencies(group).items():
                        can = candidates.get(identifier, None)
                        if can is None:
                            continue
                        updated_req = can.req

                        if req.is_named:
                            version = next(s.version for s in updated_req.specifier)
                            if options.save_strategy == "minimum":
                                req.specifier = get_specifier(f">={version}")
                            elif options.save_strategy == "compatible":
                                req.specifier = get_specifier(f"~={version}")
                            elif options.save_strategy == "exact":
                                req.specifier = get_specifier(f"=={version}")
                            if isinstance(req, CondaRequirement) and isinstance(updated_req, CondaRequirement):
                                req.is_python_package = updated_req.is_python_package

                            group_requirements[identifier] = req
                        elif options.save_strategy == "exact":
                            group_requirements[identifier] = updated_req
                    if group_requirements:
                        requirements[(group, i >= len(groups))] = group_requirements

                num_groups = len(requirements)
                for i, ((group, is_dev), group_requirements) in enumerate(requirements.items()):
                    project.add_dependencies(
                        group_requirements,
                        to_group=group,
                        dev=is_dev,
                        show_message=i == num_groups - 1,
                    )



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.venv.list import ListCommand as BaseCommand
from pdm.cli.commands.venv.utils import get_venv_prefix
from pdm.exceptions import NoPythonVersion
from pdm.termui import _console, _err_console

if TYPE_CHECKING:
    import argparse
    from pdm.project import Project


class ListCommand(BaseCommand):
    description = BaseCommand.__doc__

    def _get_venvs(self, project, options):
        with _console.capture() as capture:
            super().handle(project, options)
        return [venv for venv in capture.get().splitlines() if venv]

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm_conda.project import CondaProject

        project = cast(CondaProject, project)
        venvs = self._get_venvs(project, options)

        # add conda venvs
        if project.conda_config.is_initialized:
            conda_venvs = []
            try:
                with _console.capture(), _err_console.capture():
                    project_interpreter = project.resolve_interpreter()
                if (venv := project_interpreter.get_venv()) and venv.is_conda:
                    venv_prefix = get_venv_prefix(project)
                    ident = venv.root.name
                    if venv_prefix in ident:
                        ident = ident[len(venv_prefix) :]
                    conda_venvs.append(f"*  [success]{ident}[/]: {venv.root}")
            except NoPythonVersion:
                pass

            with project.conda_config.with_conda_venv_location() as (_, overriden):
                if overriden:
                    conda_venvs += self._get_venvs(project, options)

            for venv in conda_venvs:
                if venv not in venvs:
                    venvs.append(venv)

        for venv in venvs:
            project.core.ui.echo(venv)



from __future__ import annotations

import contextlib
import json
import subprocess
from functools import lru_cache
from pathlib import Path
from shutil import which
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from pdm.cli.commands.venv.backends import VirtualenvCreateError
from pdm.exceptions import (
    InstallationError,
    PdmException,
    RequirementError,
    UninstallError,
)
from pdm.models.finder import ReverseVersion
from pdm.models.setup import Setup
from pdm.termui import Verbosity

from pdm_conda import logger
from pdm_conda.models.candidates import CondaCandidate, parse_channel
from pdm_conda.models.conda import ChannelSorter
from pdm_conda.models.config import CondaRunner
from pdm_conda.models.requirements import (
    CondaRequirement,
    parse_conda_version,
    parse_requirement,
)
from pdm_conda.models.setup import CondaSetupDistribution
from pdm_conda.utils import normalize_name

if TYPE_CHECKING:
    from typing import Iterable

    from pdm_conda.project import CondaProject


class CondaExecutionError(PdmException):
    pass


class CondaResolutionError(CondaExecutionError):
    pass


class CondaSearchError(CondaExecutionError):
    pass


class CondaRunnerNotFoundError(CondaExecutionError):
    pass


@contextlib.contextmanager
def _optional_temporary_file(environment: dict | list):
    """
    If environment contains data then creates temporary file else yield None
    :param environment: environment data
    :return: Temporary file or None
    """
    if environment:
        with NamedTemporaryFile(mode="w+", suffix=".yml" if isinstance(environment, dict) else ".lock") as f:
            yield f
    else:
        yield


def run_conda(
    cmd,
    exception_cls: type[PdmException] = CondaExecutionError,
    exception_msg: str = "Error locking dependencies",
    **environment,
) -> dict:
    """
    Optionally creates temporary environment file and run conda command
    :param cmd: conda command
    :param exception_cls: exception to raise on error
    :param exception_msg: base message to show on error
    :param environment: environment or lockfile data
    :return: conda command response
    """
    executable = which(cmd[0])
    if executable is None:
        raise CondaRunnerNotFoundError(f"Conda runner {cmd[0]} not found.")

    lockfile = environment.get("lockfile", [])
    with _optional_temporary_file(lockfile or environment) as f:
        if lockfile or environment:
            if lockfile:
                f.write("\n".join(lockfile))
            elif environment:
                for name, options in environment.items():
                    if options:
                        f.write(f"{name}:")
                        if isinstance(options, str):
                            f.write(f" {options}\n")
                            continue
                        else:
                            f.write("\n")
                        for v in options:
                            f.write(f"  - {v}\n")
            f.seek(0)
            cmd += ["--file", f.name]
        logger.debug(f"cmd: {' '.join(cmd)}")
        if environment:
            logger.debug(f"env: {environment}")
        process = subprocess.run(cmd, capture_output=True, encoding="utf-8")
    if "--json" in cmd:
        try:
            response = json.loads(process.stdout)
        except:
            response = {}
    else:
        response = {}
        msg = f"{process.stdout}\n" if process.stdout else ""
        if process.stderr:
            msg += process.stderr
        if msg:
            response["message"] = msg
    try:
        process.check_returncode()
    except subprocess.CalledProcessError as e:
        msg = ""
        if exception_msg:
            msg = f"{exception_msg}\n"
        if isinstance(response, dict) and not response.get("success", False):
            if err := response.get("solver_problems", response.get("error", response.get("message", process.stderr))):
                if isinstance(err, str):
                    err = [err]
                msg += "\n".join(err)
        else:
            msg += process.stderr
        if environment:
            msg += f"\n{environment}"
        raise exception_cls(msg) from e
    return response


@lru_cache(maxsize=None)
def _get_channel_sorter(platform: str, channels: tuple[str]) -> ChannelSorter:
    """
    Get channel sorter
    :param channels: list of conda channels used to determine priority
    :param platform: env platform
    :return: channel sorter
    """
    return ChannelSorter(platform, channels)


def sort_candidates(
    project: CondaProject,
    packages: list[CondaCandidate],
    minimal_version: bool,
) -> Iterable[CondaCandidate]:
    """
    Sort candidates following mamba specification
    (https://mamba.readthedocs.io/en/latest/advanced_usage/package_resolution.html).
    :param project: PDM project
    :param packages: list of conda candidates
    :return: sorted conda candidates
    """
    if len(packages) <= 1:
        return packages
    channels_sorter = _get_channel_sorter(project.platform, tuple(project.conda_config.channels))

    def get_preference(candidate: CondaCandidate):
        return (
            not candidate.track_feature,
            ReverseVersion(candidate.version) if minimal_version else candidate.version,
            candidate.build_number,
            -channels_sorter.get_priority(candidate.channel or ""),
            candidate.timestamp,
        )

    return sorted(packages, key=get_preference, reverse=True)


def _parse_candidates(project: CondaProject, packages: list[dict], requirement=None) -> list[CondaCandidate]:
    """
    Convert conda packages to candidates
    :param project: PDM project
    :param packages: conda packages
    :param requirement: requirement linked to packages
    :return: list of candidates
    """
    candidates = []
    for p in packages:
        dependencies = p.get("depends", None) or []
        valid_candidate = True
        for d in dependencies:
            if d.startswith("__"):
                d = parse_requirement(f"conda:{d}")
                if not any(d.is_compatible(v) for v in project.virtual_packages):
                    valid_candidate = False
                    break
        if valid_candidate:
            candidates.append(CondaCandidate.from_conda_package(p, requirement))

    return candidates


def _ensure_channels(
    project: CondaProject,
    channels: list[str],
    log_message: str = "No channels specified, using defaults if exist.",
) -> list[str]:
    """
    Ensure channels and if empty use defaults
    :param project: PDM project
    :param channels: channels to validate
    :param log_message: log message to display if using defaults
    :return: list of channels
    """
    channels = channels or project.conda_config.channels
    if not channels:
        project.core.ui.echo(log_message, verbosity=Verbosity.DEBUG)
        channels.append("defaults")

    return list(dict.fromkeys(channels))


@lru_cache(maxsize=None)
def _conda_search(
    project: CondaProject,
    requirement: str,
    channels: tuple[str],
) -> list[dict]:
    """
    Search conda candidates for a requirement
    :param project: PDM project
    :param requirement: requirement
    :param channels: requirement channels
    :return: list of conda candidates
    """
    config = project.conda_config
    command = config.command("search")
    command.append(requirement)

    for c in channels:
        command.extend(["-c", c])
    if channels:
        command.append("--override-channels")
    command.append("--json")
    try:
        result = run_conda(command)
    except RequirementError as e:
        if "PackagesNotFoundError:" in str(e):
            result = dict()
        else:
            raise

    if config.runner == CondaRunner.CONDA:
        packages = result.get(parse_requirement(f"conda:{requirement}").name, [])
    else:
        packages = result.get("result", dict()).get("pkgs", [])
    return packages


def conda_search(
    project: CondaProject,
    requirement: CondaRequirement | str,
    channel: str | None = None,
) -> list[CondaCandidate]:
    """
    Search conda candidates for a requirement
    :param project: PDM project
    :param requirement: requirement
    :param channel: requirement channel
    :return: list of conda candidates
    """
    _requirement = requirement
    if isinstance(_requirement, CondaRequirement):
        channel = channel or _requirement.channel
        _requirement = _requirement.as_line(with_build_string=True, conda_compatible=True).replace(" ", "=")
    if "::" in _requirement:
        channel, _requirement = _requirement.split("::", maxsplit=1)
    if isinstance(requirement, str):
        requirement = parse_requirement(f"conda:{requirement}")
    channels = _ensure_channels(
        project,
        [channel] if channel else [],
        f"No channel specified for searching [req]{requirement}[/] using defaults if exist.",
    )
    packages = _conda_search(project, _requirement, tuple(channels))
    return _parse_candidates(project, packages, requirement)


def conda_create(
    project: CondaProject,
    requirements: list[CondaRequirement],
    channels: list[str] | None = None,
    prefix: Path | str | None = None,
    name: str = "",
    dry_run: bool = False,
    fetch_candidates: bool = True,
) -> dict[str, list[CondaCandidate]]:
    """
    Creates environment using conda
    :param project: PDM project
    :param requirements: conda requirements
    :param channels: requirement channels
    :param prefix: environment prefix
    :param name: environment name
    :param dry_run: don't install if dry run
    :param fetch_candidates: if True ensure ensure candidates were fetched
    """
    config = project.conda_config
    if not config.is_initialized:
        raise VirtualenvCreateError("Error creating environment, no pdm-conda configs were found on pyproject.toml.")
    candidates = dict()
    channels = channels or []
    for req in requirements:
        if req.channel:
            channels.append(req.channel)
    channels = _ensure_channels(
        project,
        channels,
        "No channels specified for creating environment, using defaults if exist.",
    )
    command = config.command("create")
    command.append("--json")
    if prefix is not None:
        command.extend(["--prefix", str(prefix)])
    elif name:
        command.extend(["--name", name])
    else:
        raise VirtualenvCreateError("Error creating environment, name or prefix must be specified.")

    if dry_run:
        command.append("--dry-run")

    command += list(
        {
            req.as_line(with_build_string=True, conda_compatible=True, with_channel=True).replace(" ", "="): None
            for req in requirements
        },
    )

    if channels:
        for c in channels:
            command.extend(["-c", c])
        command.append("--override-channels")

    result = run_conda(
        command,
        exception_cls=CondaResolutionError if dry_run else VirtualenvCreateError,
        exception_msg=f"Error resolving requirements with {config.runner}" if dry_run else "Error creating environment",
    )
    if fetch_candidates:
        actions = result.get("actions", dict())
        fetch_packages = {pkg["name"]: pkg for pkg in actions.get("FETCH", [])}
        packages = actions.get("LINK", [])
        for i, pkg in enumerate(packages):
            pkg = fetch_packages.get(pkg["name"], pkg)
            if any(True for n in ("constrains", "depends") if n not in pkg):
                pkg = conda_search(
                    project,
                    f'{pkg["name"]}={pkg["version"]}={pkg["build_string"]}',
                    parse_channel(pkg["channel"]),
                )
            packages[i] = pkg

        _requirements = {req.conda_name: req for req in requirements}
        for pkg in packages:
            # if is list of candidates then it comes from search
            if isinstance(pkg, list):
                if pkg:
                    candidates[pkg[0].name] = pkg
            else:
                name = pkg["name"]
                candidates[name] = _parse_candidates(
                    project,
                    packages=[pkg],
                    requirement=_requirements.get(name, None),
                )
    return candidates


def conda_env_remove(project: CondaProject, prefix: Path | str | None = None, name: str = "", dry_run: bool = False):
    """
    Removes environment using conda
    :param project: PDM project
    :param prefix: environment prefix
    :param name: environment name
    :param dry_run: don't install if dry run
    """
    config = project.conda_config
    if not config.is_initialized:
        raise VirtualenvCreateError("Error removing environment, no pdm-conda configs were found on pyproject.toml.")
    command = config.command("env remove")
    command.append("--json")
    if prefix is not None:
        command += ["--prefix", str(prefix)]
    elif name:
        command += ["--name", name]
    else:
        raise VirtualenvCreateError("Error removing environment, name or prefix must be specified.")

    if dry_run:
        command.append("--dry-run")

    run_conda(command, exception_cls=VirtualenvCreateError, exception_msg="Error removing environment")


def conda_env_list(project: CondaProject) -> list[Path]:
    """
    List Conda environments
    :param project: PDM project
    :return: list of conda environments
    """
    config = project.conda_config
    command = config.command("env list")
    command.append("--json")
    environments = run_conda(command, exception_cls=CondaExecutionError, exception_msg="Error listing environments")
    return [Path(env) for env in environments.get("envs", [])]


def _conda_install(
    command: list[str],
    packages: str | list[str] | None = None,
    exception_cls: type[PdmException] = InstallationError,
    dry_run: bool = False,
    explicit: bool = False,
):
    if isinstance(packages, str):
        packages = [packages]
    if not packages:
        raise exception_cls(f"No packages used for {' '.join(command[:3])} command.")
    if explicit:
        packages.insert(0, "@EXPLICIT")
    else:
        command.extend(packages)
    if dry_run:
        command.append("--dry-run")
    kwargs: dict = dict()
    if explicit:
        kwargs["lockfile"] = packages
    run_conda(command + ["--json"], exception_cls=exception_cls, exception_msg="", **kwargs)


def conda_install(
    project: CondaProject,
    packages: str | list[str],
    dry_run: bool = False,
    no_deps: bool = False,
):
    """
    Install resolved packages using conda
    :param project: PDM project
    :param packages: resolved packages
    :param dry_run: don't install if dry run
    :param no_deps: don't install dependencies if true
    """
    config = project.conda_config
    command = config.command("install")
    if no_deps:
        command.append("--no-deps")
        if config.runner != CondaRunner.MICROMAMBA:
            command.append("--no-update-deps")

    if config.installation_method == "copy":
        _copy = "copy"
        if config.runner == CondaRunner.MICROMAMBA:
            _copy = f"always-{_copy}"
        command.append(f"--{_copy}")

    _conda_install(command, packages, dry_run=dry_run, explicit=True)


def conda_uninstall(
    project: CondaProject,
    packages: str | list[str],
    dry_run: bool = False,
    no_deps: bool = False,
):
    """
    Uninstall resolved packages using conda
    :param project: PDM project
    :param packages: resolved packages
    :param dry_run: don't uninstall if dry run
    :param no_deps: don't uninstall dependencies if true
    """
    config = project.conda_config

    command = config.command("remove")
    if no_deps:
        if config.runner == CondaRunner.MICROMAMBA:
            command.append("--no-prune")
        command.append("--force")

    _conda_install(command, packages, dry_run=dry_run, exception_cls=UninstallError)


def not_initialized_warning(project):
    project.core.ui.echo(
        "[warning]Tried to execute a conda command but no pdm-conda configs were found on pyproject.toml.[/]",
    )


def conda_info(project: CondaProject) -> dict:
    """
    Get conda info containing virtual packages, default channels and packages
    :param project: PDM project
    :return: dict with conda info
    """
    config = project.conda_config
    res: dict = dict(virtual_packages=set(), platform="", channels=[])
    if config.is_initialized:
        info = run_conda(config.command("info") + ["--json"])
        if config.runner != CondaRunner.MICROMAMBA:
            virtual_packages = {"=".join(p) for p in info["virtual_pkgs"]}
        else:
            virtual_packages = set(info["virtual packages"])

        res["virtual_packages"] = {parse_requirement(f"conda:{p.replace('=', '==', 1)}") for p in virtual_packages}
        res["platform"] = info["platform"]
        res["channels"] = [parse_channel(channel) for channel in (info["channels"] or [])]
    else:
        not_initialized_warning(project)
    return res


def conda_list(project: CondaProject) -> dict[str, CondaSetupDistribution]:
    """
    List conda installed packages
    :param project: PDM project
    :return: packages distribution
    """
    config = project.conda_config
    distributions = dict()
    if config.is_initialized:
        packages = run_conda(config.command("list") + ["--json"])
        for package in packages:
            if config.runner != CondaRunner.MICROMAMBA and package.get("platform", "") == "pypi":
                continue
            name, version = package["name"], package["version"]
            distributions[normalize_name(name)] = CondaSetupDistribution(
                Setup(
                    name=name,
                    summary="",
                    version=parse_conda_version(version),
                ),
                package=package,
            )
    else:
        not_initialized_warning(project)
    return distributions



import functools
from collections import ChainMap
from copy import copy

from pdm.environments import PythonEnvironment

from pdm_conda.mapping import pypi_to_conda
from pdm_conda.project import CondaProject
from pdm_conda.utils import normalize_name

_patched = False


def wrap_get_working_set(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        dist_map = {
            normalize_name(pypi_to_conda(dist.metadata["Name"])): dist for dist in getattr(res, "_dist_map").values()
        }
        setattr(res, "_dist_map", dist_map)
        setattr(res, "_iter_map", ChainMap(dist_map, getattr(res, "_shared_map", {})))
        return res

    return wrapper


def wrap_init(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        res = func(self, *args, **kwargs)
        if isinstance(self.project, CondaProject):
            self.project = copy(self.project)
            self.project.environment = self
        return res

    return wrapper


if not _patched:
    setattr(PythonEnvironment, "__init__", wrap_init(PythonEnvironment.__init__))
    setattr(PythonEnvironment, "get_working_set", wrap_get_working_set(PythonEnvironment.get_working_set))
    _patched = True



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.cli.commands.init import Command as BaseCommand
from pdm.cli.options import ArgumentGroup

from pdm_conda.project import CondaProject

if TYPE_CHECKING:
    import argparse

    from pdm_conda.project import Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "init"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        conda_group = ArgumentGroup("Conda Options")
        conda_group.add_argument(
            "-cr",
            "--runner",
            dest="conda_runner",
            type=str,
            help="Specify Conda runner executable",
            default="",
        )
        conda_group.add_argument(
            "-c",
            "--channel",
            dest="conda_channel",
            type=str,
            help="Specify Conda channel",
            default="",
        )
        conda_group.add_to_parser(parser)

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        config = project.conda_config
        overriden = dict()
        if runner := options.conda_runner:
            config.runner = runner
            overriden["runner"] = runner
            config.is_initialized = True
            if (channel := options.conda_channel) and channel not in config.channels:
                config.channels.append(channel)
                overriden["channels"] = config.channels

        super().handle(project, options)
        if runner:
            with config.force_set_project_config():
                for key, value in overriden.items():
                    setattr(config, key, value)
            project.pyproject.write(show_message=False)
